import re

from ..util import LookAhead
from .segmenter import Segmenter
from .segments import MatchableSegment, Segment

WHITESPACE = set(["whitespace", "break"])
PARAGRAPH_END = set(["break"])
SENTENCE_END = set(["period", "epoint", "qmark"])
MIN_SENTENCE = 3


class ParagraphsSentencesAndWhitespace(Segmenter):
    """
    Constructs a paragraphs, sentences and whitespace segmenter.  This segmenter
    is intended to be used in western languages where sentences and paragraphs
    are meaningful segments of text content.

    Tree structure:

    * whitespace : :class:`~deltas.Segment`
    * paragraph : :class:`~deltas.MatchableSegment`
      * sentence : :class:`~deltas.MatchableSegment`
      * whitespace : :class:`~deltas.Segment`

    :Example:
        >>> from deltas import ParagraphsSentencesAndWhitespace, text_split
        >>> from deltas.segmenters import print_tree
        >>>
        >>> a = text_split.tokenize("This comes first.  This comes second.")
        >>>
        >>> segmenter = ParagraphsSentencesAndWhitespace()
        >>> segments = segmenter.segment(a)
        >>>
        >>> print_tree(segments)
        Segment: 'This comes first.  This comes second.'
        	MatchableSegment: 'This comes first.  This comes second.'
        		MatchableSegment: 'This comes first.'
        		Segment: '  '
        		MatchableSegment: 'This comes second.'

    :Parameters:
        whitespace : `set` ( `str` )
            A set of token types that represent whitespace.
        paragraph_end : `set` ( `str` )
            A set of token types that represent the end of a pragraph.
        sentence_end : `set` ( `str`)
            A set of tokens types that represent the end of a sentence.
        min_sentence : `int`
            The minimum non-whitespace tokens that a sentence must contain
            before a sentence_end will be entertained.
    """  # noqa
    def __init__(self, *, whitespace=None,
                          paragraph_end=None,
                          sentence_end=None,
                          min_sentence=None):

        self.whitespace = set(whitespace or WHITESPACE)
        self.paragraph_end = set(paragraph_end or PARAGRAPH_END)
        self.sentence_end = set(sentence_end or SENTENCE_END)
        self.min_sentence = int(min_sentence or MIN_SENTENCE)

    def segment(self, tokens):
        """
        Segments a sequence of tokens into a sequence of segments.

        :Parameters:
            tokens : `list` ( :class:`~deltas.Token` )
        """
        look_ahead = LookAhead(tokens)

        segments = Segment()

        while not look_ahead.empty():

            if look_ahead.peek().type not in self.whitespace: # Paragraph!
                paragraph = MatchableSegment(look_ahead.i)

                while not look_ahead.empty() and \
                      look_ahead.peek().type not in self.paragraph_end:

                    if look_ahead.peek().type not in self.whitespace: #Sentence!
                        sentence = MatchableSegment(look_ahead.i,
                                                    [next(look_ahead)])

                        while not look_ahead.empty() and \
                              look_ahead.peek().type not in self.paragraph_end:

                            sentence.append(next(look_ahead))

                            if sentence[-1].type in self.sentence_end:
                                non_whitespace = sum(s.type not in self.whitespace for s in sentence)
                                if non_whitespace >= self.min_sentence:
                                    break

                        paragraph.append(sentence)

                    else:  # look_ahead.peek().type in self.whitespace
                        whitespace = Segment(look_ahead.i, [next(look_ahead)])
                        paragraph.append(whitespace)

                segments.append(paragraph)
            else: # look_ahead.peek().type in self.whitespace
                whitespace = Segment(look_ahead.i, [next(look_ahead)])
                segments.append(whitespace)


        return segments

    @classmethod
    def from_config(cls, config, name, section_key="segmenters"):
        section = config[section_key][name]
        return cls(
            whitespace=section.get('whitespace'),
            paragraph_end=section.get('paragraph_end'),
            sentence_end=section.get('sentence_end'),
            min_sentence=section.get('min_sentence')
        )



import re
from collections import namedtuple

import yamlconf

from .token import Token


class Tokenizer:
    """
    Constructs a tokenizaton strategy.
    """
    def tokenize(self, text, token_class=Token):
        """
        Tokenizes a text.
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="tokenizers"):
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        else:
            Tokenizer = yamlconf.import_module(section['class'])
            return Tokenizer.from_config(config, name, section_key)


class RegexTokenizer(Tokenizer):
    """
    Uses a lexicon of regular expressions and names to tokenize a text string.
    """
    def __init__(self, lexicon):
        self.lexicon = lexicon
        self.regex = re.compile('|'.join('(?P<{0}>{1})'.format(name, pattern)
                                         for name, pattern in lexicon))

    def tokenize(self, text, token_class=None):
        return [t for t in self._tokenize(text, token_class=token_class)]

    def _tokenize(self, text, token_class=None):
        """
        Tokenizes a text

        :Returns:
            A `list` of tokens
        """
        token_class = token_class or Token
        tokens = {}

        for i, match in enumerate(self.regex.finditer(text)):
            value = match.group(0)

            try:
                token = tokens[value]
            except KeyError:
                type = match.lastgroup
                token = token_class(value, type=type)
                tokens[value] = token

            yield token



from nose.tools import eq_

from ..util import LookAhead

def test_lookahead_list():
    
    l = [1,2,3]
    
    look_ahead = LookAhead(l)
    eq_(look_ahead.peek(), 1)
    eq_(look_ahead.pop(), 1)
    eq_(look_ahead.peek(), 2)
    
    for i, val in enumerate(look_ahead):
        eq_(val, i+2)
    

def test_lookahead_generator():
    
    l = (i for i in [1,2,3])
    
    look_ahead = LookAhead(l)
    eq_(look_ahead.peek(), 1)
    eq_(look_ahead.pop(), 1)
    eq_(look_ahead.peek(), 2)
    
    for i, val in enumerate(look_ahead):
        eq_(val, i+2)

def test_lookahead_identity():
    
    l1 = LookAhead([])
    l2 = LookAhead(l1)
    
    eq_(l1, l2)
    eq_(id(l1), id(l2)) # Confirms same memory location



import re

from nose.tools import eq_

from ...tokenizers import Token, wikitext_split
from .. \
    paragraphs_sentences_and_whitespace import ParagraphsSentencesAndWhitespace
from ..segments import MatchableSegment, Segment


def test_segment():
    
    segmenter = ParagraphsSentencesAndWhitespace()
    
    text = 'This is some text.  This is some other text.\n ' + \
           'A. Peterson is a name that I made up.\n' + \
           'This is an additional sentence.\n' + \
           '\n' + \
           '== OMG HEADER ==\n' + \
           'This is a new paragraph! Isn\'t this fun?'
    
    tokens = wikitext_split.tokenize(text)
    segments = list(segmenter.segment(tokens))
    print([str(s) for s in segments[0]])
    
    eq_(len(segments), 3) # 2 paragraphs + 1 whitespace
    
    eq_(len(segments[0]), 7) #4 sentences + 3 whitespace
    
    eq_(type(segments[1]), Segment)
    eq_(type(segments[2]), MatchableSegment)
    eq_(type(segments[2][2]), MatchableSegment)
    
    eq_(''.join(str(s) for s in segments), text)



from nose.tools import eq_

from ..wikitext_split import wikitext_split


def test_wikitext_split():

    input = "As a sentence, this 34 includes punctuation. \n" + \
            "\n" + \
            "==Header!==\n" + \
            "克·科伊尔 しメfoobar!" + \
            "And then we have another sentence here!\n" + \
            "https://website.gov?param=value\n" + \
            "peoples' ain't d’encyclopédie\n" + \
            "mailto:email@email.mail"

    expecteds = ['As', ' ', 'a', ' ', 'sentence', ',', ' ', 'this', ' ',
                 '34', ' ', 'includes', ' ', 'punctuation', '.', ' \n\n',
                 '==', 'Header', '!', '==', '\n',
                 '克', '·', '科', '伊', '尔', ' ', 'し', 'メ', 'foobar', '!',
                 'And', ' ', 'then', ' ', 'we', ' ', 'have', ' ', 'another',
                 ' ', 'sentence',
                 ' ', 'here', '!', '\n', 'https://website.gov?param=value', '\n',
                 'peoples\'', ' ', 'ain\'t', ' ', 'd’encyclopédie', '\n',
                 'mailto:email@email.mail']

    tokens = list(wikitext_split.tokenize(input))

    for token, expected in zip(tokens, expecteds):
        eq_(token, expected)



from .tokenizer import RegexTokenizer

LEXICON = [
    ('word',          r'[^[\W\d]+'),
    ('number',        r'[\d]+'),
    ('period',        r'\.+'),
    ('qmark',         r'\?+'),
    ('epoint',        r'!+'),
    ('comma',         r',+'),
    ('colon',         r':+'),
    ('scolon',        r';+'),
    ('break',         r'(\n|\n\r|\r\n)\s*(\n|\n\r|\r\n)+'),
    ('whitespace',    r'[\n\r\s]+'),
    ("etc",           r".")
]

text_split = RegexTokenizer(LEXICON)



import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def requirements(fname):
    return [line.strip()
            for line in open(os.path.join(os.path.dirname(__file__), fname))]

setup(
    name = "deltas",
    version = "0.3.7",
    author = "Aaron Halfaker",
    author_email = "ahalfaker@wikimedia.org",
    description = "An experimental diff library for generating " + \
                  "operation deltas that represent the " + \
                  "difference between two sequences of comparable items.",
    license = "MIT",
    url = "https://github.com/halfak/Deltas",
    packages=find_packages(),
    long_description = read('README.rst'),
    install_requires = ['yamlconf'],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: General",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering"
    ]
)
